"""

This module contains source finding and photometry routines.

"""

import astropy.io.fits as pyfits
import astropy.table as atpy
import astropy.constants as constants
import numpy as np
import numpy.fft as fft
from astLib import *
import os
import pylab
import math
from scipy import ndimage
from scipy import interpolate
from . import catalogs
from . import maps
from . import completeness
from . import signals
import sys

#------------------------------------------------------------------------------------------------------------
def findObjects(filteredMapDict, threshold = 3.0, minObjPix = 3, rejectBorder = 10, 
                findCenterOfMass = True, removeRings = True, ringThresholdSigma = 0, invertMap = False, 
                objIdent = 'ACT-CL', longNames = False, verbose = True, useInterpolator = True, 
                measureShapes = False, DS9RegionsPath = None):
    """Finds objects in the filtered maps pointed to by the imageDict. Threshold is in units of sigma 
    (as we're using S/N images to detect objects). Catalogs get added to the imageDict.
        
    Throw out objects within rejectBorder pixels of edge of frame if rejectBorder != None
    
    Set invertMap == True to do a test of estimating fraction of spurious sources
    
    Set measureShapes == True to fit ellipses in a similar style to how SExtractor does it (may be useful for
    automating finding of extended sources).
    
    This now find rings around extremely bright sources and adds them to the survey mask if removeRings is True
    
    """
           
    if rejectBorder is None:
        rejectBorder=0

    data=filteredMapDict['SNMap']
    areaMask=filteredMapDict['surveyMask']
    wcs=filteredMapDict['wcs']
    flagMask=filteredMapDict['flagMask']
    
    # This is for checking contamination
    if invertMap == True:
        data=data*-1
        
    # Thresholding to identify significant pixels
    # Objects in rings will be discarded; we update the survey area mask accordingly in-place here
    objIDs, objPositions, objNumPix, segMap=getObjectPositions(data, threshold, 
                                                               findCenterOfMass = findCenterOfMass)
    if removeRings == True:
        minRingPix=30
        ringIDs, ringPositions, ringNumPix, ringSegMap=getObjectPositions(data, ringThresholdSigma, 
                                                                          findCenterOfMass = True)    
        for i in range(len(ringIDs)):
            if type(ringNumPix) != float and ringNumPix[i] > minRingPix:
                y, x=ringPositions[i]
                if ringSegMap[int(y), int(x)] != ringIDs[i]:
                    ringSegMap[np.equal(ringSegMap, ringIDs[i])]=-1*ringSegMap[np.equal(ringSegMap, ringIDs[i])]
        ringSegMap=np.array(np.less(ringSegMap, 0), dtype = int)
        ringMask=ringSegMap
        #filteredMapDict['flagMask']=filteredMapDict['flagMask']-ringMask
    else:
        ringMask=None

    # In the past this had problems - Simone using happily in his fork though, so now optional
    if useInterpolator == True:
        mapInterpolator=interpolate.RectBivariateSpline(np.arange(data.shape[0]), 
                                                        np.arange(data.shape[1]), 
                                                        data, kx = 3, ky = 3)
                                                    
    # Border around edge where we might throw stuff out just to cut down contamination
    if type(areaMask) == np.ndarray and areaMask.sum() > 0:
        minX=np.where(areaMask > 0)[1].min()
        maxX=np.where(areaMask > 0)[1].max()
        minY=np.where(areaMask > 0)[0].min()
        maxY=np.where(areaMask > 0)[0].max()
    else:
        minX=0
        maxX=segMap.shape[1]-1
        minY=0
        maxY=segMap.shape[0]-1
    minX=minX+rejectBorder
    maxX=maxX-rejectBorder
    minY=minY+rejectBorder
    maxY=maxY-rejectBorder
        
    # Build catalog - initially as a list of dictionaries, then converted to astropy Table
    catalog=[]
    idNumCount=1
    for i in range(len(objIDs)):
        if type(objNumPix) != float and objNumPix[i] > minObjPix:
            objDict={}
            objDict['id']=idNumCount
            objDict['x']=objPositions[i][1]
            objDict['y']=objPositions[i][0]
            if ringMask is not None:
                if ringMask[int(objDict['y']), int(objDict['x'])] > 0:
                    continue
            objDict['RADeg'], objDict['decDeg']=wcs.pix2wcs(objDict['x'], objDict['y'])
            if objDict['RADeg'] < 0:
                objDict['RADeg']=360.0+objDict['RADeg']
            galLong, galLat=astCoords.convertCoords("J2000", "GALACTIC", objDict['RADeg'], objDict['decDeg'], 2000)
            objDict['galacticLatDeg']=galLat
            if longNames == False:
                objDict['name']=catalogs.makeName(objDict['RADeg'], objDict['decDeg'], prefix = objIdent)
            else:
                objDict['name']=catalogs.makeLongName(objDict['RADeg'], objDict['decDeg'], prefix = objIdent)                    
            objDict['numSigPix']=objNumPix[i]
            objDict['template']=filteredMapDict['label']
            objDict['tileName']=filteredMapDict['tileName']
            if useInterpolator == True:
                objDict['SNR']=mapInterpolator(objDict['y'], objDict['x'])[0][0]
            else:
                objDict['SNR']=data[int(round(objDict['y'])), int(round(objDict['x']))]
            objDict['flags']=flagMask[int(round(objDict['y'])), int(round(objDict['x']))]
            # Optional SExtractor style shape measurements                
            if measureShapes == True:
                doubleCheck=False
                if objDict['numSigPix'] > 9:
                    mask=np.equal(segMap, objIDs[i])
                    ys, xs=np.where(segMap == objIDs[i])
                    yMin=ys.min()
                    xMin=xs.min()
                    yMax=ys.max()
                    xMax=xs.max()
                    # Centres (1st order moments) - cx2, cy2 here to avoid overwriting whatever catalog cx, cy is (see above)
                    xs=xs-xMin
                    ys=ys-yMin
                    cx2=(xs*data[mask]).sum()/data[mask].sum()
                    cy2=(ys*data[mask]).sum()/data[mask].sum()
                    # Spread (2nd order moments)
                    x2=(((xs**2)*data[mask]).sum()/data[mask].sum())-cx2**2
                    y2=(((ys**2)*data[mask]).sum()/data[mask].sum())-cy2**2
                    xy=(((xs*ys)*data[mask]).sum()/data[mask].sum())-cx2*cy2
                    # A, B, theta from above - correct theta has same sign as xy (see SExtractor manual) 
                    theta=np.degrees(np.arctan(2*(xy/(x2-y2)))/2.0)
                    if xy > 0 and theta < 0:
                        theta=theta+90
                    elif xy < 0 and theta > 0:
                        theta=theta-90
                    if theta > 0 and xy > 0:
                        doubleCheck=True
                    if theta < 0 and xy < 0:
                        doubleCheck=True
                    if doubleCheck == True:
                        A=np.sqrt((x2+y2)/2.0 + np.sqrt( ((x2-y2)/2)**2 + xy**2))
                        B=np.sqrt((x2+y2)/2.0 - np.sqrt( ((x2-y2)/2)**2 + xy**2))
                        # Moments work terribly for low surface brightness, diffuse things which aren't strongly peaked
                        # Shape measurement is okay though - so just scale A, B to match segMap area
                        segArea=float(np.count_nonzero(np.equal(segMap, objIDs[i])))
                        curArea=A*B*np.pi
                        scaleFactor=np.sqrt(segArea/curArea)
                        A=A*scaleFactor
                        B=B*scaleFactor  
                        ecc=np.sqrt(1-B**2/A**2)
                        objDict['ellipse_PA']=theta
                        objDict['ellipse_A']=A
                        objDict['ellipse_B']=B
                        objDict['ellipse_x0']=cx2+xMin
                        objDict['ellipse_y0']=cy2+yMin
                        objDict['ellipse_e']=ecc
                if objDict['numSigPix'] <= 9 or doubleCheck == False:
                    objDict['ellipse_PA']=-99
                    objDict['ellipse_A']=-99
                    objDict['ellipse_B']=-99
                    objDict['ellipse_x0']=-99
                    objDict['ellipse_y0']=-99
                    objDict['ellipse_e']=-99
            # Add to catalog (masks now applied before we get here so no need to check)
            if objDict['SNR'] > threshold:
                catalog.append(objDict)
            idNumCount=idNumCount+1     
    
    # From here on, catalogs should be astropy Table objects...
    if len(catalog) > 0:
        catalog=catalogs.catalogListToTab(catalog)
        if DS9RegionsPath is not None:
            catalogs.catalog2DS9(catalog, DS9RegionsPath)

    return catalog

#------------------------------------------------------------------------------------------------------------
def getObjectPositions(mapData, threshold, findCenterOfMass = True):
    """Creates a segmentation map and find objects above the given threshold.
    
    Args:
        mapData (:obj:`numpy.ndarray`): The 2d map to segment.
        threshold (float): The threshold above which objects will be selected.
        findCenterOfMass: If True, return the object center weighted according to the values in mapData. If
            False, return the pixel that holds the maximum value.
    Returns:
        objIDs (:obj:`numpy.ndarray`): Array of object ID numbers.
        objPositions (list): List of corresponding (y, x) positions.
        objNumPix (:obj:`numpy.ndarray`): Array listing number of pixels per object.
        segmentationMap (:obj:`numpy.ndarray`): The segmentation map (2d array).
    
    """
    
    if threshold < 0:
        raise Exception("Detection threshold (thresholdSigma in the config file) cannot be negative unless in forced photometry mode.")
            
    sigPix=np.array(np.greater(mapData, threshold), dtype=int)
    sigPixMask=np.equal(sigPix, 1)    
    segmentationMap, numObjects=ndimage.label(sigPix)
    objIDs=np.unique(segmentationMap)
    if findCenterOfMass == True:
        objPositions=ndimage.center_of_mass(mapData, labels = segmentationMap, index = objIDs)
    else:
        objPositions=ndimage.maximum_position(mapData, labels = segmentationMap, index = objIDs)
    objNumPix=ndimage.sum(sigPixMask, labels = segmentationMap, index = objIDs)

    return objIDs, objPositions, objNumPix, segmentationMap

#------------------------------------------------------------------------------------------------------------
def getSNRValues(catalog, SNMap, wcs, useInterpolator = True, invertMap = False, prefix = ''):
    """Measures SNR values in given map at catalog positions.
    
    Set invertMap == True, to do a test for estimating the spurious source fraction
    
    Use prefix to add prefix before SNR in catalog (e.g., fixed_SNR)
    
    Use template to select a particular filtered map (i.e., label of mapFilter given in .par file, e.g.,
    a filter used to measure properties for all objects at fixed scale).
    
    """
            
    # This is for checking contamination
    if invertMap == True:
        SNMap=SNMap*-1
        
    # In the past this had problems - Simone using happily in his fork though, so now optional
    if useInterpolator == True:
        mapInterpolator=interpolate.RectBivariateSpline(np.arange(SNMap.shape[0]), 
                                                        np.arange(SNMap.shape[1]), 
                                                        SNMap, kx = 3, ky = 3)
    
    if len(catalog) > 0:
        catalog.add_column(atpy.Column(np.zeros(len(catalog)), prefix+'SNR'))
    for obj in catalog:
        x, y=wcs.wcs2pix(obj['RADeg'], obj['decDeg'])
        if int(x) > 0 and int(x) < SNMap.shape[1] and int(y) > 0 and int(y) < SNMap.shape[0]:
            if useInterpolator == True:
                obj[prefix+'SNR']=mapInterpolator(y, x)[0][0]
            else:
                obj[prefix+'SNR']=SNMap[int(round(y)), int(round(x))] # read directly off of S/N map
           
#------------------------------------------------------------------------------------------------------------
def measureFluxes(catalog, filteredMapDict, diagnosticsDir, photFilteredMapDict = None,
                  useInterpolator = True, ycObsFreqGHz = 148.0):
    """Add flux measurements to each catalog pointed to in the imageDict. Measured in 'outputUnits' 
    specified in the filter definition in the .par file (and written to the image header as 'BUNIT').
    
    Maps should have been normalised before this stage to make the value of the filtered map at the object
    position equal to the flux that is wanted (yc, uK or Jy/beam). This includes taking out the effect of
    the beam.
    
    Use 'photFilter' to choose which filtered map in which to make flux measurements at fixed scale
    (fixed_delta_T_c, fixed_y_c etc.). Set to None if you don't want these.
    
    If the filtered map is in yc, then columns that contain the amplitude in in uK CMB delta temperature
    will be added, named deltaT_c, assuming that the observing frequency is as set by ycObsFreqGHz

    """
        
    mapData=filteredMapDict['data']
    wcs=filteredMapDict['wcs']         
    mapUnits=wcs.header['BUNIT']
    
    # Adds fixed_SNR values to catalogs for all maps
    if photFilteredMapDict is not None:
        getSNRValues(catalog, photFilteredMapDict['SNMap'], wcs, prefix = 'fixed_', 
                     useInterpolator = useInterpolator)
    
    # These keywords would only be written if output units 'uK' (point source finding)
    # This also relies on the beam solid angle being in the header of the beam .txt file
    if 'BEAMNSR' in wcs.header.keys():
        beamSolidAngle_nsr=wcs.header['BEAMNSR']
        obsFreqGHz=wcs.header['FREQGHZ']
        reportJyFluxes=True
    else:
        reportJyFluxes=False
    
    # In the past this had problems - Simone using happily in his fork though, so now optional
    if useInterpolator == True:
        mapInterpolator=interpolate.RectBivariateSpline(np.arange(mapData.shape[0]), 
                                                        np.arange(mapData.shape[1]), 
                                                        mapData, kx = 3, ky = 3)
    else:
        mapInterpolator=None
    
    # Add fixed filter scale maps
    mapDataList=[mapData]
    interpolatorList=[mapInterpolator]
    prefixList=['']
    if photFilteredMapDict is not None:
        photMapData=photFilteredMapDict['data']
        mapDataList.append(photMapData)
        prefixList.append('fixed_')
        if useInterpolator == True:
            photMapInterpolator=interpolate.RectBivariateSpline(np.arange(photMapData.shape[0]), 
                                                                np.arange(photMapData.shape[1]), 
                                                                photMapData, kx = 3, ky = 3)
        else:
            photMapInterpolator=None
        interpolatorList.append(photMapInterpolator)

    keysToAdd=['deltaT_c', 'err_deltaT_c']
    if mapUnits == 'yc':
        keysToAdd=keysToAdd+['y_c', 'err_y_c']
    elif mapUnits == 'uK':
        keysToAdd=keysToAdd+['fluxJy', 'err_fluxJy']        
    for prefix in prefixList:
        for k in keysToAdd:
            if len(catalog) > 0:
                catalog.add_column(atpy.Column(np.zeros(len(catalog)), prefix+k))

    for obj in catalog:
        x, y=wcs.wcs2pix(obj['RADeg'], obj['decDeg'])
        for data, prefix, interpolator in zip(mapDataList, prefixList, interpolatorList):
            # NOTE: We might want to avoid 2d interpolation here because that was found not to be robust elsewhere
            # 2018: Simone seems to now be using this happily, so now optional
            if useInterpolator == True:
                mapValue=interpolator(y, x)[0][0]
            else:
                mapValue=data[int(round(y)), int(round(x))]
            # NOTE: remember, all normalisation should be done when constructing the filtered maps, i.e., not here!
            if mapUnits == 'yc':
                yc=mapValue
                obj[prefix+'y_c']=yc/1e-4                            # So that same units as H13 in output catalogs
                obj[prefix+'err_y_c']=obj[prefix+'y_c']/obj[prefix+'SNR']
                deltaTc=maps.convertToDeltaT(yc, obsFrequencyGHz = ycObsFreqGHz)
                obj[prefix+'deltaT_c']=deltaTc
                obj[prefix+'err_deltaT_c']=abs(deltaTc/obj[prefix+'SNR'])
            elif mapUnits == 'uK':
                # For this, we want deltaTc to be source amplitude
                deltaTc=mapValue
                obj[prefix+'deltaT_c']=deltaTc
                obj[prefix+'err_deltaT_c']=deltaTc/obj[prefix+'SNR']                        
                if reportJyFluxes == True:
                    obj[prefix+"fluxJy"]=deltaTToJyPerSr(obj[prefix+'deltaT_c'], obsFreqGHz)*beamSolidAngle_nsr*1.e-9
                    obj[prefix+"err_fluxJy"]=deltaTToJyPerSr(obj[prefix+'err_deltaT_c'], obsFreqGHz)*beamSolidAngle_nsr*1.e-9

#------------------------------------------------------------------------------------------------------------
def makeForcedPhotometryCatalog(filteredMapDict, inputCatalog, useInterpolator = True,\
                                DS9RegionsPath = None):
    """Make a catalog on which we can do forced photometry at the locations of objects in it.
    
    inputCatalog is either a path or a table object.
        
    """
    
    if type(inputCatalog) == str:
        forcedTab=atpy.Table().read(inputCatalog)
    elif type(inputCatalog) == atpy.Table:
        forcedTab=inputCatalog
    else:
        raise Exception("Data type for inputCatalog should be a path, or an astropy Table object")
    RAKey, decKey=catalogs.getTableRADecKeys(forcedTab)
    mask=np.less(forcedTab[RAKey], 0)
    forcedTab[RAKey][mask]=360-abs(forcedTab[RAKey][mask])
    forcedTab.rename_column(RAKey, 'RADeg')
    forcedTab.rename_column(decKey, 'decDeg')
    if 'name' not in forcedTab.keys():
        forcedTab.add_column(atpy.Column(np.arange(len(forcedTab))+1, 'name'))
    
    wcs=filteredMapDict['wcs']
    areaMask=filteredMapDict['surveyMask'] 
    tileName=filteredMapDict['tileName']
    data=filteredMapDict['SNMap']
    if useInterpolator == True:
        mapInterpolator=interpolate.RectBivariateSpline(np.arange(data.shape[0]), 
                                                        np.arange(data.shape[1]), 
                                                        data, kx = 3, ky = 3)

    forcedTab=catalogs.getCatalogWithinImage(forcedTab, data.shape, wcs)
    catalog=[]
    idNumCount=1
    for row in forcedTab:
        objDict={}
        objDict['id']=idNumCount
        x, y=wcs.wcs2pix(row['RADeg'], row['decDeg'])
        x, y=int(round(x)), int(round(y))
        if data[y, x] != 0:
            objDict['x']=x
            objDict['y']=y
            objDict['RADeg'], objDict['decDeg']=row['RADeg'], row['decDeg']
            galLong, galLat=astCoords.convertCoords("J2000", "GALACTIC", objDict['RADeg'], objDict['decDeg'], 2000)
            objDict['galacticLatDeg']=galLat
            objDict['name']=row['name']
            objDict['numSigPix']=1
            objDict['template']=filteredMapDict['label']
            objDict['tileName']=filteredMapDict['tileName']
            if useInterpolator == True:
                objDict['SNR']=mapInterpolator(objDict['y'], objDict['x'])[0][0]
            else:
                objDict['SNR']=data[int(round(objDict['y'])), int(round(objDict['x']))]
            catalog.append(objDict)
            idNumCount=idNumCount+1

    # From here on, catalogs should be astropy Table objects...
    if len(catalog) > 0:
        catalog=catalogs.catalogListToTab(catalog)
        if DS9RegionsPath is not None:
            catalogs.catalog2DS9(catalog, DS9RegionsPath)
            
    return catalog

#------------------------------------------------------------------------------------------------------------
def addFreqWeightsToCatalog(imageDict, photFilter, diagnosticsDir):
    """Add relative weighting by frequency for each object in the optimal catalog, extracted from the 
    data cube saved under diagnosticsDir (this is made by makeSZMap in RealSpaceMatchedFilter). This is
    needed for multi-frequency cluster finding / analysis - for, e.g., weighting relativistic corrections to
    y0~ (which was estimated from the inverse variance weighted average of y0~ from each frequency map).
    
    NOTE: this is only applied for the reference filter (pointed to by photFilter)
    
    """

    if photFilter is None:
        return None
    
    catalog=imageDict['optimalCatalog']
    for tileName in imageDict['tileNames']:
        label=photFilter+"#"+tileName
        freqWeightMapFileName=diagnosticsDir+os.path.sep+"freqRelativeWeights_%s.fits" % (label)
        if os.path.exists(freqWeightMapFileName) == True:
            img=pyfits.open(freqWeightMapFileName)
            freqCube=img[0].data
            wcs=astWCS.WCS(img[0].header, mode = 'pyfits')
            obsFreqGHzDict={}
            for key in wcs.header.keys():
                if key.find("FREQGHZ") != -1:
                    obsFreqGHzDict[int(key.split("FREQGHZ")[0])]=wcs.header[key]
            for i in range(len(obsFreqGHzDict.keys())):
                freqStr=("%.1f" % (obsFreqGHzDict[i])).replace(".", "p")    # MongoDB doesn't like '.' in key names
                keyToAdd='fixed_y_c_weight_%sGHz' % (freqStr)
                if keyToAdd not in catalog.keys():
                    catalog.add_column(atpy.Column(np.zeros(len(catalog)), keyToAdd))
            for row in catalog:
                if row['template'].split("#")[-1] == tileName:
                    x, y=wcs.wcs2pix(row['RADeg'], row['decDeg'])
                    x=int(round(x))
                    y=int(round(y))
                    freqWeights=freqCube[:, y, x]
                    for i in range(len(obsFreqGHzDict.keys())):
                        freqStr=("%.1f" % (obsFreqGHzDict[i])).replace(".", "p")    # MongoDB doesn't like '.' in key names
                        row['fixed_y_c_weight_%sGHz' % (freqStr)]=freqWeights[i]

#------------------------------------------------------------------------------------------------------------
def deltaTToJyPerSr(temp, obsFreqGHz):
    """Convert delta T (uK) to Jy/sr at the given frequency in GHz
    
    """
    
    kB=constants.k_B.cgs.value
    h=constants.h.cgs.value
    c=constants.c.cgs.value
    nu=obsFreqGHz*1.e9
    T0=signals.TCMB
    x=h*nu/(kB*T0)
    cNu=2*(kB*T0)**3/(h**2*c**2)*x**4/(4*(np.sinh(x/2.))**2)
    cNu*=1e23
    
    return temp*cNu*1e-6/T0

#------------------------------------------------------------------------------------------------------------
def JyPerSrToDeltaT(JySr, obsFreqGHz):
    """Convert Jy/sr to delta T (uK) at the given frequency in GHz
    
    """
    
    kB=constants.k_B.cgs.value
    h=constants.h.cgs.value
    c=constants.c.cgs.value
    nu=obsFreqGHz*1.e9
    T0=signals.TCMB
    x=h*nu/(kB*T0)
    cNu=2*(kB*T0)**3/(h**2*c**2)*x**4/(4*(np.sinh(x/2.))**2)
    cNu*=1e23
    
    temp_uK=(JySr*T0)/(cNu*1e-6)
    
    return temp_uK

#------------------------------------------------------------------------------------------------------------
def getRadialDistanceMap(objDict, data, wcs):
    """Returns an array of same dimensions as data but containing radial distance to the object specified
    in the objDict in units degrees on the sky
    
    """
    
    x0=objDict['x']
    y0=objDict['y']
    x1=x0+1
    y1=y0+1
    ra1, dec1=wcs.pix2wcs(x1, y1)
    xPixScale=astCoords.calcAngSepDeg(objDict['RADeg'], objDict['decDeg'], ra1, objDict['decDeg'])
    yPixScale=astCoords.calcAngSepDeg(objDict['RADeg'], objDict['decDeg'], objDict['RADeg'], dec1)       
    xRange=np.array([np.arange(0, data.shape[1])-x0]*data.shape[0])*xPixScale
    yRange=(np.array([np.arange(0, data.shape[0])-y0]*data.shape[1])*yPixScale).transpose()
    rRange=np.sqrt(xRange**2+yRange**2)       

    return rRange

#------------------------------------------------------------------------------------------------------------
def getPixelsDistanceMap(objDict, data):
    """Returns an array of same dimensions as data but containing radial distance to the object specified
    in the objDict in units of pixels
    
    """
    
    x0=objDict['x']
    y0=objDict['y']
    x1=x0+1
    y1=y0+1     
    xRange=np.array([np.arange(0, data.shape[1])-x0]*data.shape[0])
    yRange=(np.array([np.arange(0, data.shape[0])-y0]*data.shape[1])).transpose()
    rRange=np.sqrt(xRange**2+yRange**2)       

    return rRange

#------------------------------------------------------------------------------------------------------------
def makeAnnulus(innerScalePix, outerScalePix):
    """Makes the footprint of an annulus for feeding into the rank filter
    
    Returns annulus, numValues
    
    """

    # Make the footprint for the rank filter
    #bckInnerScalePix=beamSigmaPix*3#int(round((bckInnerRadiusArcmin/60.0)/yPixScale))
    #bckOuterScalePix=beamSigmaPix*5#int(round((bckOuterRadiusArcmin/60.0)/yPixScale))
    bckInnerScalePix=int(round(innerScalePix))
    bckOuterScalePix=int(round(outerScalePix))
    annulus=np.zeros([bckOuterScalePix*2, bckOuterScalePix*2])
    xRange=np.array([np.arange(annulus.shape[1])-bckOuterScalePix]*annulus.shape[0])
    yRange=np.array([np.arange(annulus.shape[0])-bckOuterScalePix]*annulus.shape[1]).transpose()
    rRange=np.sqrt(xRange**2+yRange**2)
    annulus[np.logical_and(np.greater(rRange, bckInnerScalePix),
                                np.less(rRange, bckOuterScalePix))]=1.0
    #numValues=annulus.flatten().nonzero()[0].shape[0]
    
    return annulus.astype('int64')




"""

This module contains tools for handling catalogs, which are usually :obj:`astropy.table.Table` objects.

"""

from astLib import *
import numpy as np
import operator
import os
import sys
import time
import astropy.table as atpy
from astropy.coordinates import SkyCoord
from astropy.coordinates import match_coordinates_sky
import astropy.io.fits as pyfits
from scipy import ndimage
from . import maps

# For adding meta data to output
import datetime
import nemo

#------------------------------------------------------------------------------------------------------------
XMATCH_RADIUS_DEG=1.4/60.0  # catalog matching radius, for sim comparisons

#------------------------------------------------------------------------------------------------------------
# Definitions of table column names and formats that can be written in catalogs
COLUMN_NAMES    = ['name', 
                   'RADeg', 
                   'decDeg', 
                   'SNR', 
                   'numSigPix', 
                   'template', 
                   'tileName',
                   'flags',
                   'galacticLatDeg',
                   'deltaT_c',
                   'err_deltaT_c',
                   'y_c',
                   'err_y_c',
                   'Y500_sr',
                   'err_Y500_sr',
                   'fluxJy',
                   'err_fluxJy',
                   'redshift',
                   'redshiftErr',
                   'ellipse_PA',
                   'ellipse_A',
                   'ellipse_B',
                   'ellipse_x0',
                   'ellipse_y0',
                   'ellipse_e'
                   ]
COLUMN_FORMATS  = ['%s',
                   '%.6f',
                   '%.6f',
                   '%.1f',
                   '%d',
                   '%s',
                   '%s',
                   '%d',
                   '%.6f',
                   '%.3f',
                   '%.3f',
                   '%.3e',
                   '%.3e',
                   '%.3e',
                   '%.3e',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   '%.3f',
                   ]

columnsToAdd=[]
formatsToAdd=[]
for k in COLUMN_NAMES:
    if k in ['y_c', 'err_y_c', 'deltaT_c', 'err_deltaT_c']:
        i=COLUMN_NAMES.index(k)
        formatsToAdd.append(COLUMN_FORMATS[i])
        columnsToAdd.append("fixed_"+k)
COLUMN_NAMES=COLUMN_NAMES+columnsToAdd
COLUMN_FORMATS=COLUMN_FORMATS+formatsToAdd
        
if len(COLUMN_NAMES) != len(COLUMN_FORMATS):
    raise Exception("COLUMN_NAMES and COLUMN_FORMATS lists should be same length")

#------------------------------------------------------------------------------------------------------------
def _posRecFitFunc(snr, snrFold, pedestal, norm):
    """Fitting function used for position recovery offset (') in terms of fixed_SNR - see
    positionRecovery/positionRecoveryTestDriver.py.
    
    NOTE: Don't use this directly - call checkCrossMatch instead.
    
    """
    return norm*np.exp(-snr/snrFold)+pedestal
    
#------------------------------------------------------------------------------------------------------------
def checkCrossMatch(distArcmin, fixedSNR, z = None, addRMpc = 0.5, fitSNRFold = 1.164, fitPedestal = 0.685,
                    fitNorm = 38.097):
    """Checks the cross match offset between a cluster detection and an external catalog using a model derived
    from source injection sims (see :func:`nemo.maps.positionRecoveryAnalysis`). The position recovery test
    itself only accounts for the effect of noise fluctuations in the maps on the recovered SZ cluster
    positions.
    
    Args:
        distArcmin (:obj:`bool`): Distance of the potential cross match from the ACT position in arcmin.
        fixed_SNR (:obj:`float`): Signal-to-noise at reference filter scale (fixed_SNR) in ACT catalog.
        z (:obj:`float`, optional): If given, addRMpc will be converted to arcmin at this redshift, and then added
            in quadrature to the cross matching radius from the position recovery model.
        addRMpc (:obj:`float`, optional): Accounts for additional positional uncertainty (probably unknown) 
            in the external cross match catalog. This will be added in quadrature.
        fitSNRFold (:obj:`float`, optional): Model fit parameter - e-folding 
            (see :func:`nemo.maps.positionRecoveryAnalysis`).
        fitPedestal (:obj:`float`, optional): Model fit parameter - pedestal level
            (see :func:`nemo.maps.positionRecoveryAnalysis`).
        fitNorm (:obj:`float`, optional): Model fit parameter - normalization
            (see :func:`nemo.maps.positionRecoveryAnalysis`).
    
    Returns:
        True if distArcmin < model offset (+ optional addRMpc in arcmin at z), False if not.

    Note:
        The default values for the fit parameters are from a run on the f090, f150 ACT DR5 co-added maps
        (as used in the `ACT DR5 cluster catalog paper <https://ui.adsabs.harvard.edu/abs/2020arXiv200911043H/abstract>`_),
        and describe a function that recovers 99.7% of the inserted model clusters in 
        source injection simulations.
        
    """
    
    maxRadiusArcmin=_posRecFitFunc(fixedSNR, fitSNRFold, fitPedestal, fitNorm)
    addArcmin=0.0
    if z is not None and z > 0:
        addArcmin=np.degrees(addRMpc/astCalc.da(z))*60.0
    maxRadiusArcmin=np.sqrt(maxRadiusArcmin**2 + addArcmin**2)
    if distArcmin < maxRadiusArcmin:
        return True
    else:
        return False

#------------------------------------------------------------------------------------------------------------
def makeOptimalCatalog(catalogDict, constraintsList = []):
    """Identifies common objects between every catalog in the input dictionary of catalogs, and creates a 
    master catalog with one entry per object, keeping only the details of the highest signal-to-noise 
    detection.
    
    Args:
        catalogDict (:obj:`dict`): Dictionary where each key points to a catalog of objects.
        constraintsList (:obj:`list`, optional): A list of constraints (for the format, see 
            :func:`selectFromCatalog`).
        
    Returns:
        None - an ``optimalCatalog`` key is added to ``catalogDict`` in place.
    
    """
    
    allCatalogs=[]
    for key in catalogDict.keys():
        if len(catalogDict[key]['catalog']) > 0:
            allCatalogs.append(catalogDict[key]['catalog'])
    if len(allCatalogs) > 0:
        allCatalogs=atpy.vstack(allCatalogs)
        mergedCatalog=allCatalogs.copy()
        mergedCatalog.add_column(atpy.Column(np.zeros(len(mergedCatalog)), 'toRemove'))
        for row in allCatalogs:
            rDeg=astCoords.calcAngSepDeg(row['RADeg'], row['decDeg'], allCatalogs['RADeg'].data, 
                                         allCatalogs['decDeg'].data) 
            xIndices=np.where(rDeg < XMATCH_RADIUS_DEG)[0]
            if len(xIndices) > 1:
                xMatches=allCatalogs[xIndices]
                xMatchIndex=np.argmax(xMatches['SNR'])
                for index in xIndices:
                    if index != xIndices[xMatchIndex]:
                        mergedCatalog['toRemove'][index]=1
        mergedCatalog=mergedCatalog[mergedCatalog['toRemove'] == 0]
        mergedCatalog.remove_column('toRemove')
        mergedCatalog.sort(['RADeg', 'decDeg'])
        mergedCatalog=selectFromCatalog(mergedCatalog, constraintsList)
    else:
        mergedCatalog=atpy.Table(names=('SNR', 'RADeg', 'decDeg'))
    
    return mergedCatalog

#------------------------------------------------------------------------------------------------------------
def catalog2DS9(catalog, outFileName, constraintsList = [], addInfo = [], idKeyToUse = 'name', 
                RAKeyToUse = 'RADeg', decKeyToUse = 'decDeg', color = "cyan", showNames = True,
                writeNemoInfo = True, coordSys = 'fk5', regionShape = 'point', width = 1):
    """Writes a DS9 region file corresponding to the given catalog. 
    
    Args:
        catalog (:obj:`astropy.table.Table`): An astropy Table where each row represents an object.
        outFileName (:obj:`str`): A file name for the output DS9 region file.
        constraintsList (:obj:`list`, optional): A list of constraints in the same format as used by 
            :func:`selectFromCatalog`.
        addInfo (:obj:`list`, optional): A list of dictionaries with keys named `key` and `fmt` (e.g., 
            ``{'key': "SNR", 'fmt': "%.3f"}``). These will be added to the object label shown in DS9.
        idKeyToUse (:obj:`str`, optional): The name of the key in each object dictionary that defines the 
            object's name. Used to label objects in the DS9 region file.
        RAKeyToUse (:obj:`str`, optional): The name of the key in each object dictionary that contains the 
            RA of the object in decimal degrees.
        decKeyToUse (:obj:`str`, optional): The name of the key in each object dictionary that contains the
            declination of the object in decimal degrees.
        color (:obj:`str`, optional): The color of the plot symbol used by DS9.
        writeNemoInfo (:obj:`bool`, optional): If ``True``, writes a line with the `nemo` version and date 
            generated at the top of the DS9 .reg file.
        coordSys (:obj:`str`, optional): A string defining the coordinate system used for RA, dec, as 
            understood by DS9.
        
    Returns:
        None
    
    """
    
    cutCatalog=selectFromCatalog(catalog, constraintsList) 
    
    with open(outFileName, "w") as outFile:
        timeStamp=datetime.datetime.today().date().isoformat()
        comment="# DS9 region file"
        if writeNemoInfo == True:
            comment=comment+" generated by nemo (version: %s on %s)\n" % (nemo.__version__, timeStamp)
        else:
            comment=comment+"\n"
        outFile.write(comment)
        outFile.write('global dashlist=8 3 width=%d font="helvetica 10 normal" select=1 highlite=1 dash=0 fixed=0 edit=1 move=1 delete=1 include=1 source=1\n' % (width))
        for obj in cutCatalog:
            if len(addInfo) > 0:
                infoString=""
                for d in addInfo:
                    if infoString != "":
                        infoString=infoString+" "
                    if obj[d['key']] != None:
                        infoString=infoString+d['fmt'] % (obj[d['key']])
                    else:
                        infoString=infoString+"%s" % (str(obj[d['key']]))
                infoString=" ["+infoString+"]"
            else:
                infoString=""
            if color == 'key':
                colorString=obj['color']
            else:
                colorString=color
            if showNames == True:
                infoString=str(obj[idKeyToUse])+infoString
            if regionShape == 'point':
                outFile.write("%s;point(%.6f,%.6f) # point=cross color={%s} text={%s}\n" \
                            % (coordSys, obj[RAKeyToUse], obj[decKeyToUse], colorString, infoString))
            elif regionShape == 'circle':
                outFile.write('%s;circle(%.6f,%.6f,360") # color={%s} text={%s}\n' \
                            % (coordSys, obj[RAKeyToUse], obj[decKeyToUse], colorString, infoString))                

#------------------------------------------------------------------------------------------------------------
def makeName(RADeg, decDeg, prefix = 'ACT-CL'):
    """Makes an object name string from the given object coordinates, following the IAU convention.
    
    Args:
        RADeg (:obj:`float`): Right ascension of the object in J2000 decimal degrees.
        decDeg (:obj:`float`): Declination of the object in J2000 decimal degrees.
        prefix (:obj:`str`, optional): Prefix for the object name.
    
    Returns:
        Object name string in the format `prefix JHHMM.m+/-DDMM`.
    
    """
    
    actName=prefix+" J"+_makeRA(RADeg)+_makeDec(decDeg)
    
    return actName

#------------------------------------------------------------------------------------------------------------
def makeLongName(RADeg, decDeg, prefix = "ACT-CL"):
    """Makes a long format object name string from the given object coordinates, following the IAU convention.
    
    Args:
        RADeg (:obj:`float`): Right ascension of the object in J2000 decimal degrees.
        decDeg (:obj:`float`): Declination of the object in J2000 decimal degrees.
        prefix (:obj:`str`, optional): Prefix for the object name.
    
    Returns:
        Object name string in the format `prefix JHHMMSS.s+/-DDMMSS`.
    
    """
    
    actName=prefix+" J"+_makeLongRA(RADeg)+_makeLongDec(decDeg)
    
    return actName
    
#------------------------------------------------------------------------------------------------------------
def _makeRA(myRADeg):
    """Makes RA part of ACT names.
    
    """
    hours=(myRADeg/360)*24
    strHours=("%.10f" % (hours))
    if hours<10:
        sHours="0"+strHours[0]
    else:
        sHours=strHours[:2]
    
    mins=float(strHours[strHours.index("."):])*60
    strMins=("%.10f" % (mins))
    if mins < 10:
        sMins="0"+strMins[:3]
    else:
        sMins=strMins[:4]

    return (sHours+sMins)#[:-2] # Trims off .x as not used in ACT names
        
#------------------------------------------------------------------------------------------------------------
def _makeDec(myDecDeg):
    """Makes dec part of ACT names
    
    """
    
    # Positive
    if myDecDeg>0:
        if myDecDeg<10:
            sDeg="0"+str(myDecDeg)[0]
        else:
            sDeg=str(myDecDeg)[:2]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
        
        return "+"+sDeg+sMins
    else:
        if myDecDeg>-10:
            sDeg="-0"+str(myDecDeg)[1]
        else:
            sDeg=str(myDecDeg)[:3]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
        
        return str(sDeg+sMins)

#-------------------------------------------------------------------------------------------------------------
def _makeLongRA(myRADeg):
    """Make a long RA string, i.e. in style of long XCS names
    
    """
    
    hours=(myRADeg/360)*24
    if hours<10:
        sHours="0"+str(hours)[0]
    else:
        sHours=str(hours)[:2]
    
    mins=float(str(hours)[str(hours).index("."):])*60
    if mins<10:
        sMins="0"+str(mins)[0]
    else:
        sMins=str(mins)[:2]
        
    secs=float(str(mins)[str(mins).index("."):])*60
    if secs<10:
        sSecs="0"+str(secs)[:3]
    else:
        sSecs=str(secs)[:4]     
        
    return sHours+sMins+sSecs
        
#-------------------------------------------------------------------------------------------------------------
def _makeLongDec(myDecDeg):
    """Make a long dec sting i.e. in style of long XCS names
    
    """
    # Positive
    if myDecDeg>0:
        if myDecDeg<10:
            sDeg="0"+str(myDecDeg)[0]
        else:
            sDeg=str(myDecDeg)[:2]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
            
        secs=float(str(mins)[str(mins).index("."):])*60
        if secs<10:
            sSecs="0"+str(secs)[:3]
        else:
            sSecs=str(secs)[:4]         
        
        return "+"+sDeg+sMins+sSecs
    else:
        if myDecDeg>-10:
            sDeg="-0"+str(myDecDeg)[1]
        else:
            sDeg=str(myDecDeg)[:3]
    
        mins=float(str(myDecDeg)[str(myDecDeg).index("."):])*60
        if mins<10:
            sMins="0"+str(mins)[:1]
        else:
            sMins=str(mins)[:2]
            
        secs=float(str(mins)[str(mins).index("."):])*60
        if secs<10:
            sSecs="0"+str(secs)[:3]
        else:
            sSecs=str(secs)[:4]         
        
        return sDeg+sMins+sSecs
        
#-------------------------------------------------------------------------------------------------------------
def selectFromCatalog(catalog, constraintsList):
    """Return a table of objects matching the given constraints from the catalog. 
    
    Args:
        catalog (:obj:`astropy.table.Table`): The catalog from which objects will be selected.
        constraintsList (:obj:`list`): A list of constraints, where each item is a string of the form
            "key < value", "key > value", etc.. Note that the spaces between the key, operator 
            (e.g. '<'), and value are essential.
    
    Returns:
        An astropy Table object.
    
    """
            
    passedConstraint=catalog
    for constraintString in constraintsList:
        key, op, value=constraintString.split()
        passedConstraint=passedConstraint[eval("passedConstraint['%s'] %s %s" % (key, op, value))]

    return passedConstraint

#------------------------------------------------------------------------------------------------------------
def catalogListToTab(catalogList, keysToWrite = COLUMN_NAMES):
    """Converts a catalog in the form of a list of dictionaries (where each dictionary holds the object
    properties) into an :obj:`astropy.table.Table` object.
    
    Args:
        catalogList (:obj:`list`): Catalog in the form of a list of dictionaries.
        keysToWrite (:obj:`list`, optional): Keys to convert into columns in the output table.
    
    Returns:
        An :obj:`astropy.table.Table` object, where each row corresponds to an object in the catalog.
    
    """
    
    availKeys=list(catalogList[0].keys())
    tab=atpy.Table()
    for key in keysToWrite:
        if key in availKeys:
            arr=[]
            for obj in catalogList:
                if obj[key] != None:
                    arr.append(obj[key])
                else:
                    arr.append(-99)
            tab.add_column(atpy.Column(arr, key))

    return tab

#------------------------------------------------------------------------------------------------------------
def tabToCatalogList(tab):
    """Converts an :obj:`astropy.table.Table` object into a list of dictionaries.
    
    Args:
        tab (:obj:`astropy.table.Table`): Catalog in the form of an astropy Table object.
    
    Returns:
        A list of dictionaries, where each dictionary represents an object in the catalog.
    
    """
    
    catalog=[]
    for row in tab:
        objDict={}
        for k in list(tab.keys()):
            objDict[k]=row[k]
        catalog.append(objDict)
    
    return catalog
    
#------------------------------------------------------------------------------------------------------------
def writeCatalog(catalog, outFileName, constraintsList = []):
    """Writes the catalog to disk, including meta data on the Nemo version used to create the catalog in the
    header.
     
    Args:
        catalog (:obj:`astropy.table.Table`): The object catalog.
        outFileName (:obj:`str`): The output filename. The format depends upon the extension given. Note
            that CSV format (resulting from using the ``.csv`` extension) is written tab-delimited. 
        constraintsList (:obj:`list`, optional): A list of constraints, where each item is a string of the
            form "key < value", "key > value", etc.. Note that the spaces between the key, operator 
            (e.g. '<'), and value are essential.
        
    Returns:
        None
        
    """

    # Deal with any blank catalogs (e.g., in blank tiles - not that we would want such things...)
    if type(catalog) == list and len(catalog) == 0:
        return None
    cutCatalog=selectFromCatalog(catalog, constraintsList)
    cutCatalog.meta['NEMOVER']=nemo.__version__
    if outFileName.split(".")[-1] == 'csv':
        cutCatalog.write(outFileName, format = 'ascii.csv', delimiter = '\t', overwrite = True)
    else:
        cutCatalog.write(outFileName, overwrite = True)

#------------------------------------------------------------------------------------------------------------
def removeDuplicates(tab):
    """Removes duplicate objects from the catalog - keeping the highest SNR detection for each duplicate. 
    This routine is used to clean up the output of MPI runs (where we have overlapping tiles).
    
    Args:
        tab (:obj:`astropy.table.Table`): The object catalog to be checked for duplicates.

    Returns:
        Table with duplicates removed (:obj:`astropy.table.Table`), the number of duplicates found, and a
        list of names for the duplicated objects.
    
    """

    if len(tab) == 1:
        return tab, 1, []
    
    # Find all duplicates
    cat=SkyCoord(ra = tab['RADeg'].data, dec = tab['decDeg'].data, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(cat, cat, nthneighbor = 2)
    mask=np.less(rDeg.value, XMATCH_RADIUS_DEG)
    noDupMask=np.greater_equal(rDeg.value, XMATCH_RADIUS_DEG)
    dupTab=tab[mask]
    noDupTab=tab[noDupMask]
    
    # All duplicates removed?
    if mask.sum() == 0:
        return tab, 0, []
    
    # Much faster
    keepMask=np.zeros(len(dupTab), dtype = bool)
    for i in range(len(dupTab)):
        # NOTE: astCoords does not like atpy.Columns sometimes...
        rDeg=astCoords.calcAngSepDeg(dupTab['RADeg'][i], dupTab['decDeg'][i], dupTab['RADeg'].data, dupTab['decDeg'].data)
        mask=np.less_equal(rDeg, XMATCH_RADIUS_DEG)
        if mask.sum() == 0:	# This ought not to be possible but catch anyway
            bestIndex=i
        else:
            indices=np.where(mask == True)[0]
            bestIndex=indices[np.equal(dupTab['SNR'][mask], dupTab['SNR'][mask].max())][0]
        keepMask[bestIndex]=True
    keepTab=dupTab[keepMask]
    
    keepTab=atpy.vstack([keepTab, noDupTab])
    keepTab.sort('RADeg')
    
    return keepTab, len(dupTab), dupTab['name']

#------------------------------------------------------------------------------------------------------------
def flagTileBoundarySplits(tab, xMatchRadiusArcmin = 2.5):
    """Flag objects that are closer than some matching radius but which appear in different tiles. These are
    potentially objects that have been de-blended across tile boundaries (in which case one entry in the
    catalog should be kept and the other discarded), but some objects in this category are genuine close
    pairs. At the moment, this is best resolved by visual inspection. This routine adds a flag column named
    `tileBoundarySplit` to the catalog to make it easy to spot these cases.
        
    Args:
        tab (:obj:`astropy.table.Table`): The object catalog to be checked.
        xMatchRadiusArcmin (float, optional): Cross-match radius in arcmin.

    Returns:
        Catalog (:obj:`astropy.table.Table`) with `tileBoundarySplit` column added - this is True for objects
        that may have been deblended across tiles, and require visual inspection.
    
    """
    
    if len(tab) == 1:
        return tab
    
    xMatchRadiusDeg=xMatchRadiusArcmin/60.
    
    # Find all potential duplicates within a given matching radius
    cat=SkyCoord(ra = tab['RADeg'].data, dec = tab['decDeg'].data, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(cat, cat, nthneighbor = 2)
    mask=np.less(rDeg.value, xMatchRadiusDeg)
    noDupMask=np.greater_equal(rDeg.value, xMatchRadiusDeg)
    dupTab=tab[mask]
    noDupTab=tab[noDupMask]
        
    # Identify pairs split across tile boundaries
    tileBoundarySplit=np.zeros(len(dupTab), dtype = bool)
    for i in range(len(dupTab)):
        # NOTE: astCoords does not like atpy.Columns sometimes...
        rDeg=astCoords.calcAngSepDeg(dupTab['RADeg'][i], dupTab['decDeg'][i], dupTab['RADeg'].data, dupTab['decDeg'].data)
        mask=np.less_equal(rDeg, xMatchRadiusDeg)
        if mask.sum() == 0:	# This ought not to be possible but catch anyway
            bestIndex=i
        else:
            indices=np.where(mask == True)[0]
            bestIndex=indices[np.equal(dupTab['SNR'][mask], dupTab['SNR'][mask].max())][0]
            if np.unique(dupTab['tileName'][indices]).shape[0] > 1:
                tileBoundarySplit[indices]=True
    dupTab['tileBoundarySplit']=tileBoundarySplit
    dupTab=dupTab[tileBoundarySplit]
    
    # Flag in the main table
    tab['tileBoundarySplit']=np.zeros(len(tab), dtype = bool)
    for row in dupTab:
        mask=(tab['name'] == row['name'])
        tab['tileBoundarySplit'][mask]=True
        
    return tab

#------------------------------------------------------------------------------------------------------------
def generateRandomSourcesCatalog(mapData, wcs, numSources, seed = None):
    """Generate a random source catalog (with amplitudes in deltaT uK), with random positions within the 
    footprint of the given map (areas where pixel values == 0 are ignored). The distribution of source 
    amplitudes is roughly similar to that seen in the 148 GHz ACT maps, but this routine should only be used
    for tests - it is not a serious attempt at simulating the real extragalactic source population.
    
    Args:
        mapData (:obj:`numpy.ndarray`): Map pixel-data, only used for determining valid area in which sources
            may be randomly placed (pixel values == 0 are ignored).
        wcs (:obj:`astWCS.WCS`): WCS corresponding to the map.
        numSources (:obj:`int`): Number of random sources to put into the output catalog.
        seed (optional, :obj:`int`): If given, generate the catalog using this random seed value. This is
            useful for generating the same realization across maps at different frequencies. The seed will
            be reset after this routine exits.
    
    Returns:
        An :obj:`astropy.table.Table` object containing the catalog.
        
    """
    
    if seed is not None:
        np.random.seed(seed)
        
    deltaT=np.random.lognormal(np.log(600), 1.1, numSources)
    ys, xs=np.where(mapData != 0)
    ys=ys+np.random.uniform(0, 1, len(ys))
    xs=xs+np.random.uniform(0, 1, len(xs))
    indices=np.random.randint(0, len(ys), numSources)
    coords=wcs.pix2wcs(xs[indices], ys[indices])
    coords=np.array(coords)
    tab=atpy.Table()
    tab.add_column(atpy.Column(np.arange(0, numSources)+1, "name"))
    tab.add_column(atpy.Column(coords[:, 0], "RADeg"))
    tab.add_column(atpy.Column(coords[:, 1], "decDeg"))
    tab.add_column(atpy.Column(deltaT, "deltaT_c"))
    
    if seed is not None:
        np.random.seed()
        
    return tab

#------------------------------------------------------------------------------------------------------------
def generateTestCatalog(config, numSourcesPerTile, amplitudeColumnName = 'fixed_y_c', 
                        amplitudeRange = [0.001, 1], amplitudeDistribution = 'linear', selFn = None,
                        avoidanceRadiusArcmin = 20.0, maskDilationPix = 0,
                        tileNames = None):
    """Generate a catalog of objects with random positions and amplitudes. This is for testing purposes - 
    see, e.g., :meth:`nemo.maps.sourceInjectionTest`.
    
    Args:
        config (:obj:`nemo.startup.NemoConfig`): Nemo configuration object.
        numSourcesPerTile (:obj:`int`): The maximum number of sources to insert into each tile. The number 
            of sources actually inserted may be less than this depending on the value of 
            `avoidanceRadiusArcmin`.
        amplitudeColumnName (:obj:`str`): Name of the column in the output catalog in which source (or
            cluster) amplitudes will be stored. Typically this should be "deltaT_c" for sources, and
            "fixed_y_c" for clusters.
        amplitudeRange (:obj:`list`): Range for the random amplitudes, in the form [minimum, maximum].
        amplitudeDistribution (:obj:`str`): Either 'linear' or 'log'.
        selFn (:obj:`nemo.completeness.SelFn`, optional): Nemo selection function object, used to access 
            area masks and coordinates info. If not given, a selFn object will be created using the info 
            in the config. Providing this saves time, as the area mask files don't have to be read from 
            disk.
        avoidanceRadiusArcmin (:obj:`float`): Minimum separation between two objects in the output catalog.
            This should be set large enough to avoid crowding and spurious cross-matching problems.
        maskDilationPix (:obj:`int`): Avoid placing objects within this many pixels of the area mask, by
            dilating the holes in the mask by this number of pixels. Avoids problems with e.g., objects
            split across the mask boundary. If not set, gives a small number of objects with larger than
            expected offsets in recovered position in :meth:`nemo.maps.sourceInjectionTest`).

    Returns:
        An :obj:`astropy.table.Table` object containing the catalog.
        
    """
    
    if selFn is None:
        selFn=completeness.SelFn(config.selFnDir, 4.0, configFileName = config.configFileName, 
                                 enableCompletenessCalc = False, setUpAreaMask = True,
                                 tileNames = tileNames)
    if tileNames is None:
        tileNames=selFn.tileNames

    RAs=[]
    decs=[]
    amps=[]
    catTileNames=[]
    for tileName in tileNames:
        mapData=selFn.areaMaskDict[tileName]
        if mapData.sum() == 0:  # Skip any empty/blank tile
            continue
        wcs=selFn.WCSDict[tileName]
        # Dilate mask to avoid putting objects close to borders or holes
        for i in range(maskDilationPix):
            mapData=1-ndimage.binary_dilation(1-mapData)
        ys, xs=np.where(mapData != 0)
        ys=ys+np.random.uniform(0, 1, len(ys))
        xs=xs+np.random.uniform(0, 1, len(xs))
        indices=np.random.randint(0, len(ys), len(ys))
        coords=wcs.pix2wcs(xs[indices], ys[indices])
        coords=np.array(coords)
        tileRAs=[]
        tileDecs=[]
        keepIndices=[]
        for i in indices:
            rai=coords[i, 0]
            deci=coords[i, 1]
            rDeg=astCoords.calcAngSepDeg(rai, deci, tileRAs, tileDecs)
            keepObj=False
            if len(rDeg) > 0:
                if rDeg.min()*60 > avoidanceRadiusArcmin:
                    keepObj=True
            else:
                keepObj=True
            if keepObj == True:
                tileRAs.append(rai)
                tileDecs.append(deci)
                keepIndices.append(i)
            if len(keepIndices) == numSourcesPerTile:
                break
        # Edge case where we have tiny but non-zero area and didn't manage to insert anything
        if len(coords) == 0:
            continue
        RAs=RAs+coords[keepIndices, 0].tolist()
        decs=decs+coords[keepIndices, 1].tolist()
        if amplitudeDistribution == 'linear':
            amp=np.random.uniform(amplitudeRange[0], amplitudeRange[1], len(keepIndices))
        elif amplitudeDistribution == 'log':
            amp=np.power(10, np.random.uniform(np.log10(amplitudeRange)[0], np.log10(amplitudeRange)[1], len(keepIndices)))
        else:
            raise Exception("Must be either 'linear' or 'log'.")
        amps=amps+amp.tolist()
        catTileNames=catTileNames+[tileName]*len(amp.tolist())
    
    tab=atpy.Table()
    tab.add_column(atpy.Column(np.arange(0, len(amps))+1, "name"))
    tab.add_column(atpy.Column(RAs, "RADeg"))
    tab.add_column(atpy.Column(decs, "decDeg"))
    tab.add_column(atpy.Column(amps, amplitudeColumnName))
    tab.add_column(atpy.Column(catTileNames, 'tileName'))

    return tab

#------------------------------------------------------------------------------------------------------------
def crossMatch(refCatalog, matchCatalog, radiusArcmin = 2.5):
    """Cross matches `matchCatalog` onto `refCatalog` for objects found within some angular radius 
    (specified in arcmin).
    
    Args:
        refCatalog (:obj:`astropy.table.Table`): The reference catalog.
        matchCatalog (:obj:`astropy.table.Table`): The catalog to match onto the reference catalog.
        radiusArcmin (:obj:`float`, optional): Cross-match radius in arcmin.
    
    Returns:
        Cross-matched reference catalog, matchCatalog, and array of angular separation in degrees, for 
        objects in common within the matching radius. The cross matched columns are sorted such that rows in
        each correspond to the matched objects.
    
    """
    
    inTab=refCatalog
    outTab=matchCatalog
    RAKey1, decKey1=getTableRADecKeys(inTab)
    RAKey2, decKey2=getTableRADecKeys(outTab)
    cat1=SkyCoord(ra = inTab[RAKey1].data, dec = inTab[decKey1].data, unit = 'deg')
    xMatchRadiusDeg=radiusArcmin/60.
    cat2=SkyCoord(ra = outTab[RAKey2].data, dec = outTab[decKey2].data, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(cat1, cat2, nthneighbor = 1)
    mask=np.less(rDeg.value, xMatchRadiusDeg)  
    matched_outTab=outTab[xIndices]
    inTab=inTab[mask]
    matched_outTab=matched_outTab[mask]
    rDeg=rDeg.value[mask]
    
    return inTab, matched_outTab, rDeg

#------------------------------------------------------------------------------------------------------------
def removeCrossMatched(refCatalog, matchCatalog, radiusArcmin = 2.5):
    """Cross matches `matchCatalog` onto `refCatalog` for objects found within some angular radius 
    (specified in arcmin), and returns `refCatalog` with the matching entries removed.
    
    Args:
        refCatalog (:obj:`astropy.table.Table`): The reference catalog.
        matchCatalog (:obj:`astropy.table.Table`): The catalog to match onto the reference catalog.
        radiusArcmin (:obj:`float`, optional): Cross-match radius in arcmin.
    
    Returns:
        Cross-matched reference catalog (:obj:`astropy.table.Table`) with matches to `matchCatalog` removed.
        
    """
        
    inTab=refCatalog
    outTab=matchCatalog
    RAKey1, decKey1=getTableRADecKeys(inTab)
    RAKey2, decKey2=getTableRADecKeys(outTab)
    cat1=SkyCoord(ra = inTab[RAKey1].data, dec = inTab[decKey1].data, unit = 'deg')
    xMatchRadiusDeg=radiusArcmin/60.
    cat2=SkyCoord(ra = outTab[RAKey2].data, dec = outTab[decKey2].data, unit = 'deg')
    xIndices, rDeg, sep3d = match_coordinates_sky(cat1, cat2, nthneighbor = 1)
    mask=np.greater(rDeg.value, xMatchRadiusDeg)  
    inTab=inTab[mask]
    
    return inTab
    
#------------------------------------------------------------------------------------------------------------
def getTableRADecKeys(tab):
    """Returns the column names in the table in which RA, dec coords are stored, after trying a few possible
    name variations.
    
    Args:
        tab (:obj:`astropy.table.Table`): The table to search.
        
    Returns:
        Name of the RA column, name of the dec. column
    
    """
    RAKeysToTry=['ra', 'RA', 'RADeg']
    decKeysToTry=['dec', 'DEC', 'decDeg', 'Dec']
    RAKey, decKey=None, None
    for key in RAKeysToTry:
        if key in tab.keys():
            RAKey=key
            break
    for key in decKeysToTry:
        if key in tab.keys():
            decKey=key
            break
    if RAKey is None or decKey is None:
        raise Exception("Couldn't identify RA, dec columns in the supplied table.")
    
    return RAKey, decKey

#------------------------------------------------------------------------------------------------------------
def getCatalogWithinImage(tab, shape, wcs, mask = None):
    """Returns the subset of the catalog with coordinates within the image defined by the given `shape`, 
    `wcs`. Optionally, a `mask` may also be applied.
    
    Args:
        tab (:obj:`astropy.table.Table`): Catalog, as an astropy Table object. Must have columns called 
            'RADeg', 'decDeg' that contain object coordinates in decimal degrees.
        shape (:obj:`list`): Shape of the array corresponding to the image / map.
        wcs (:obj:`astWCS.WCS`): WCS of the image.
        mask (optional, :obj:`np.ndarray`): Mask with same dimensions and WCS as the image. Pixels with
            value = 1 indicate valid area, and pixels with value = 0 are considered to be outside the mask.
            If this is given, the returned catalog will contain only objects in the valid area defined by
            this image mask.
    
    Returns:
        An astropy Table containing the subset of objects within the image.
    
    """
    
    xyCoords=np.array(wcs.wcs2pix(tab['RADeg'].tolist(), tab['decDeg'].tolist()))
    selected=[]
    for i in range(len(tab)):
        x, y=xyCoords[i][0], xyCoords[i][1]
        if np.isnan(x) == True or np.isnan(y) == True:
            selected.append(False)
            continue
        if x >= 0 and x < shape[1]-1 and y >= 0 and y < shape[0]-1:
            if mask is not None:
                if mask[int(round(y)), int(round(x))] == 1:
                    selected.append(True)
                else:
                    selected.append(False)
            else:
                selected.append(True)
        else:
            selected.append(False)
    
    return tab[selected]

#------------------------------------------------------------------------------------------------------------
def addFootprintColumnToCatalog(tab, label, areaMask, wcs):
    """Add `footprint_label` column to the catalog, flagging objects found within the valid area of the given
    mask.

    Args:
        tab (:obj:`astropy.table.Table`): Catalog, as an astropy Table object. Must have columns called
            'RADeg', 'decDeg' that contain object coordinates in decimal degrees.
        label (:obj:`str`): A column named `footprint_label` will be added to the catalog. Objects in the
            catalog that fall within the valid area of the given area mask will have `footprint_label`
            set to True.
        areaMask (:obj:`np.ndarray`): Mask image defining the footprint corresponding to the given WCS.
            Pixels with value = 1 indicate valid area, and pixels with value = 0 are considered to be
            outside the mask.
        wcs (:obj:`astWCS.WCS`): WCS of the area mask that defines the footprint.

    Returns:
        An astropy Table with `footprint_label` column added.

    """

    inMask=np.zeros(len(tab['RADeg'].data), dtype = bool)
    coords=wcs.wcs2pix(tab['RADeg'].data, tab['decDeg'].data)
    coords=np.array(np.round(coords), dtype = int)
    mask1=np.logical_and(coords[:, 0] >= 0, coords[:, 1] >= 0)
    mask2=np.logical_and(coords[:, 0] < areaMask.shape[1], coords[:, 1] < areaMask.shape[0])
    mask=np.logical_and(mask1, mask2)
    inMask[mask]=inMask[mask]+areaMask[coords[:, 1][mask], coords[:, 0][mask]]
    tab['footprint_%s' % (label)]=inMask

    return tab



"""

Combines components from SO sims into maps for each channel of interest.

This is first done in healpix - CAR projection is done separately.

"""

import os
import sys
import healpy
import numpy as np
import IPython

bands=['la145', 'la093', 'la225', 'la280']
bands.sort()
healpixSimsDir="../SO_mbs"
refBand='la145' # file names in components below must contain this
components={'cmb': "201905_extragalactic/4096/cmb/0000/simonsobs_cmb_uKCMB_la145_nside4096_0000.fits",
            'tsz': "201905_extragalactic/4096/tsz/0000/simonsobs_tsz_uKCMB_la145_nside4096_0000.fits",
            'cib': "201905_extragalactic/4096/cib/0000/simonsobs_cib_uKCMB_la145_nside4096_0000.fits",
            'noise': "201906_noise_no_lowell/4096/noise/0000/simonsobs_noise_uKCMB_la145_nside4096_0000.fits"}

# At the moment, including the noise can add 'diamonds' pattern (from healpix?) when CAR projected?
componentsToUse=['cmb', 'tsz', 'cib']
componentsToUse=componentsToUse+['noise']

for b in bands:
    print("Making combined map for %s" % (b))
    combinedMap=None
    for c in componentsToUse:
        fileName=healpixSimsDir+os.path.sep+components[c].replace(refBand, b)
        print("   reading map %s" % (fileName))
        if combinedMap is None:
            combinedMap=healpy.read_map(fileName)
        else:
            combinedMap=combinedMap+healpy.read_map(fileName)
    outFileName='TOnly_%s.fits' % (b)
    healpy.write_map(outFileName, combinedMap, overwrite = True) 
